      program pftsearch

C     *************************************************************************
C     *                                                                       *
C     *                     *****     pftsearch.f     *****                   *
C     *                                                                       *
C     * MODEL-BASED PROCEDURE TO DETERMINE ORIGINS, ORIENTATIONS, AND         *
C     *  MAG SCALE FACTORS OF IMAGED PARTICLES.  IT CAN BE USED TO FIND OR    *
C     *  REFINE THESE PARAMETERS.                                             *
C     *                                                                       *
C     * PROJECT CUBIC (NCOL=NROW=NSEC) 3D MAP (DEFAULT ASSUME 532 SYMMETRY    *
C     *  IN REGULARLY SPACED THETA,PHI ORIENTATIONS FOR A USER-DEFINED        *
C     *  REGION OF 3D SPACE (DEFAULT IS ONE-HALF OF THE ASYMMETRIC UNIT).     *
C     *  THESE CARTESIAN PROJECTIONS ARE THEN CONVERTED TO POLAR MAPS AND     *
C     *  POLAR FFTS.                                                          *
C     *                                                                       *
C     * LOGICAL UNITS:                                                        *
C     *           1 = REFERENCE 3D MAP (READ IF MODE.GT.0)                    *
C     *           2 = INPUT PARAMETERS DATA FILE(s)                           *
C     *           3 = PRJ DATABASE FILE (DEFAULT = PFT.PRJS) (IF MODE.LT.0)   *
C     *           4 = PFT DATABASE FILE (DEFAULT = PFT.PFTS) (IF MODE.LT.0)   *
C     *           7 = INPUT RAW IMAGE DATA                                    *
C     *           8 = STATS OUTPUT FILE (EMPFT.RES1)                          *
C     *           9 = STATS OUTPUT FILE (EMPFT.RADS)                          *
C     *          10 = STATS OUTPUT FILE (EMPFT.RES2)                          *
C     *          11 = PARAMETER OUTPUT DATA FILES (EMPFT.NEW_DAT)             *
C     *                                                                       *
C     * May 2003 - program inception. Program modified from old pft.f         *
C     *                                                                       *
C     *************************************************************************

      implicit none

      character(len=256)  :: infile

#ifdef MPI
      include 'mpif.h'
      integer :: mcw = mpi_comm_world
#endif

      include '../include/allprogs.inc'

      COMMON/PFT_COM/ MODE, BIN, VERBOSE, MAG_CEN, MAG_STEP, MAG_NUM,
     $     MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90,
     $     PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO,
     $     ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, IDIM1, IDIM2,
     $     RADIUS, DEL_ANG, NVIEW_MOD, PARAM_FNAMES,
     $     IMG_FNAMES, QUICK_OMEGA

      common/ctf_com/pixsiz, units, volts, amp_fac, focus_maj, 
     $     focus_min, ang_maj, cs_coeff

      real pixsiz(999), volts(999), amp_fac(999), focus_maj(999), 
     $     focus_min(999), ang_maj(999), cs_coeff(999)
      integer units(999)

C     ---------------------------------------------
      include '../include/pifhead.inc'
      type (pifhead_t) maphead, imghead, prjhead, pfthead
      common /pifhead/   maphead, imghead, prjhead, pfthead
      integer peak_mem1, peak_mem2a, peak_mem2b, peak_mem2, mem_method
      integer sendcount, basecount, remainder, nmax
      integer recvcounts(128), displs(128)
      integer prj_cnt, prj_size, nprjs
      double precision elapsed_time
      double precision t_proj, t_total, t_oo
      integer*8 clock1, clock2, clock3, clock4
      real*4, allocatable :: mod_ts(:), mod_ps(:)
      real*4, allocatable :: prjs(:,:,:), local_prjs(:,:,:)
      real*4, allocatable :: map3d(:,:,:), prj(:,:), prj_avg(:,:)
      real*4  mod_om, the1, the2, phi1, tempval
      integer*4 id, k
C     ---------------------------------------------

      integer nptles, err_code, ptle_count(256), i, j, nbox
      integer :: filename_len=256, pfilesmx=999

      integer inunit
      integer*4 BIN, VERBOSE, MODE, NFILES, COMPARE_MODE, QUICK_OMEGA
      integer*4 SYM, NCOL, NROW, NSEC, NROT, NANNULI, IDIM1, IDIM2
      integer*4 NVIEW_MOD, NPIX, JCUT, ANN_LO, ANN_HI
      integer*4 NCOLBIN, NROWBIN, NSECBIN
      integer*4 MAG_NUM, MAG_NORM, alloc_stat
      real*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
      real*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX, FILT_FAC(10)
      real*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
      real*4    RADIUS, SIGCUT, PI
      character(len=256) :: param_fnames(999), img_fnames(999)
      character(len=256) :: map_filename
      LOGICAL   VAROPT
      
      integer, allocatable :: indx(:)
      real, allocatable    :: pixels(:), theta(:), phi(:), omega(:)
      real, allocatable    :: x(:), y(:), mag(:)
      real, allocatable    :: pft_ccs(:), prj_ccs(:), cmp_ccs(:)
      real, allocatable    :: ctf_val(:,:), abs_ctf(:,:), disef(:,:)

#ifdef MPI
      call mpi_init(mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_init')
      endif
      call mpi_comm_rank(mcw, iam, mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_comm_rank')
      endif
      call mpi_comm_size(mcw, num_nodes, mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_comm_size')
      endif
#else
      iam = 0
      num_nodes = 1
#endif

      call system_clock(clock1)
      pi = acos(-1.0)

c     ------------------------ MASTER -------------------------------
      if(iam == 0) then
         write(*,*) 'PFTsearch: Model-based PFT analysis'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 

         call parse_cmd_line(infile)

C     Get user input
         if(len_trim(infile) == 0) then
            inunit = 5
            call key_info (ncol, nrow, nsec, ncolbin, nrowbin, nsecbin,
     $           nrot, nannuli, npix, nfiles, compare_mode, filt_fac,
     $           map_filename, inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call key_info (ncol, nrow, nsec, ncolbin, nrowbin, nsecbin,
     $           nrot, nannuli, npix, nfiles, compare_mode, filt_fac,
     $           map_filename, inunit)
            close(17)
         endif    

C     Determine numbers of particles in parameter files
         oddeven = 0
         ptle_count = 0
         call get_ptle_counts(nfiles, filename_len, param_fnames,
     $        oddeven, ptle_count)
         nptles = sum(ptle_count)

         if(nptles == 0) then
            call error_stop('PPFT', 'main program', 
     +           'All particle parameter files are empty')
         else
            write(*,*)
            write(*,*) 'Total number of images: ', nptles
            do i=1,nfiles
               write(*,*) 'file #', i, ': ', ptle_count(i), 
     $              trim(param_fnames(i))
            enddo
            write(*,*)
         endif

      endif

C     Distribute data to all processes
#ifdef MPI
      call mpi_bcast(pixsiz,       999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(volts,        999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(amp_fac,      999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(cs_coeff,     999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(focus_maj,    999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(focus_min,    999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(ang_maj,      999, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(units,        999, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(ncol,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nrow,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nsec,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(ncolbin,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nrowbin,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nsecbin,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nrot,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nannuli,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(npix,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nfiles,         1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(compare_mode,   1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(filt_fac,      10, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mode,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(bin,            1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(verbose,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(mag_cen,        1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag_step,       1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag_num,        1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(mag_norm,       1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(xcen_model,     1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(ycen_model,     1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(sym,            1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(dthe,           1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(dphi_90,        1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(pftrad_lo,      1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(pftrad_hi,      1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(jcut,           1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(pixres_min,     1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(pixres_max,     1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(ann_lo,         1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(ann_hi,         1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(varopt,         1, mpi_logical, 0, mcw, mpi_err)
      call mpi_bcast(sigcut,         1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(res_min,        1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(res_max,        1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(idim1,          1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(idim2,          1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(radius,         1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(del_ang,        1, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(nview_mod,      1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(quick_omega,    1, mpi_integer, 0, mcw, mpi_err)

      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(map_filename, 256, mpi_character, 0, mcw, mpi_err)

      call mpi_bcast(nfiles,     1,      mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles,     1,      mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(ptle_count, nfiles, mpi_integer, 0, mcw, mpi_err)
#endif

C     Allocate global and local arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles), 
     $     pft_ccs(nptles), prj_ccs(nptles), cmp_ccs(nptles))
      
C     Get particle parameters
      if(iam == 0) then
         oddeven = 0
         nbox = 1 !rss (temporary)
         call get_ptle_params(nfiles, filename_len, param_fnames,
     $        oddeven, nptles, nbox, indx, theta, phi, omega, x, y, mag)
      endif

C     Distrubute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
#endif

C     ----------------------------------------------------
      if(iam == 0) then
         tempval =  (ncolbin*nrowbin*nview_mod)
         tempval =  (4*tempval)
         if(tempval .gt. 2.0**32) then
            print *, '*************************************************'
            print *, 'warning - each process is asking for > 4 gb'
            print *, 'you''re asking for too much memory.'
            print *, '   number of cols   - ', ncolbin
            print *, '   number of rows   - ', nrowbin
            print *, '   number of views  - ', nview_mod
            print *, '   == ', tempval, ' total bytes requested.'
            print *, '          (4 bytes per pixel) '
         endif
      endif

      allocate(prj_avg(ncolbin,nrowbin), mod_ts(nview_mod),
     $     mod_ps(nview_mod), prj(ncolbin,nrowbin), stat = alloc_stat)
      if(alloc_stat .gt. 0) then
        print *, '*************************************************'
        print *, '* getmem1 memory allocation error: exit program *'
        call tsbend
      endif

C     Determine range of values for theta and phi
      call get_t1t2p1 (sym, the1, the2, phi1)

C     Get (theta,phi) pairs that cover the asymmetric units
      call calc_mod_tps (mod_ts, mod_ps, the1, the2, phi1, dthe, 
     1  dphi_90, nview_mod, sym)

c     get projections
c     mode > 0, calculate projections
c     mode == -1 or -2, use prjs stored on disk
c     note, unless mode == -1 or -2, read entire 3d map into memory 

      call system_clock(clock2)
      if(mode >= 0) then

         allocate(map3d(ncolbin,nrowbin,nsecbin), stat = alloc_stat)
         if(alloc_stat > 0) then
            print *, '**********************************************'
            print *, '* not enough memory for 3d map: exit program *'
            call tsbend
         endif
         
c     pif_read_wmapi2 not filling in edges of map3d for bin=2
c     initialize to zero to make sure that we don't get nan's
         if(iam == 0) then
            map3d = 0.0         ! f90 array syntax

            maphead%lunum = 1
            call pif_open (map_filename, maphead%lunum, 'r') 
            call pif_read_gh (maphead)
            if(maphead%mode == 20) then
               maphead%imgnum = 1 ! load up "img" #1
               call pif_read_dh(maphead)
               call pif_read_wmapi2(maphead, map3d, ncolbin, nrowbin,
     $              nsecbin)
            else
               print *, 'invalid pif mode for 3d map'
               call tsbend
            endif
            call pif_close(maphead%lunum)

         endif
         
#ifdef MPI
         call mpi_bcast(map3d, ncolbin*nrowbin*nsecbin, mpi_real,
     $        0, mcw, mpi_err)
#endif
         
         basecount = nview_mod/num_nodes
         remainder = nview_mod - basecount*num_nodes
         
         if(iam >= (num_nodes-remainder)) then
            sendcount = basecount + 1
         else
            sendcount = basecount
         endif

         recvcounts = 0
         displs = 0
         do i=0,num_nodes-1
            if(i >= (num_nodes-remainder)) then
               recvcounts(i+1) = basecount + 1
            else
               recvcounts(i+1) = basecount
            endif
         enddo
         displs(1) = 0
         do i=1,num_nodes-1
            displs(i+1) = displs(i) + recvcounts(i)
         enddo

C     Determine best scheme for minimizing peak memory usage
C     Method 1: peak mem usage = (map + full set of projections)
C     Method 2: peak mem usage = (map + local set of projections) OR
C     (local set of projections + full set of projections)

         nmax = ceiling(real(nview_mod)/real(num_nodes))

         peak_mem1  = nsecbin   + nview_mod ! (map + full set)
         peak_mem2a = nsecbin   + nmax      ! (map + local set)
         peak_mem2b = nview_mod + nmax      ! (full set + local set)
         peak_mem2 = max(peak_mem2a, peak_mem2b)
         if(peak_mem1 < peak_mem2 .or. num_nodes == 1) then
            mem_method = 1
         else
            mem_method = 2
         endif

         if(iam == 0) then
            write(*,*)
            write(*,*) 'Memory management method = ', mem_method
            write(*,*)
         endif

         if(mem_method == 1) then
            allocate(prjs(ncolbin,nrowbin,nview_mod))
            do id=1,sendcount
               k = displs(iam+1) + id
               mod_om = 90.0
               call map_prj_ppft(prj, map3d, ncolbin, nrowbin, nsecbin,
     $              mod_ts(k), mod_ps(k), mod_om, 1.0, real(nsecbin))
               prjs(:,:,k) = prj
            enddo
            deallocate(map3d)
         endif

         if(mem_method == 2) then
            allocate(local_prjs(ncolbin,nrowbin,sendcount))
            do id=1,sendcount
               k = displs(iam+1) + id
               mod_om = 90.0
               call map_prj_ppft(prj, map3d, ncolbin, nrowbin, nsecbin,
     $              mod_ts(k), mod_ps(k), mod_om, 1.0, real(nsecbin))
               local_prjs(:,:,id) = prj
            enddo
            deallocate(map3d)
            allocate(prjs(ncolbin,nrowbin,nview_mod))
            do id=1,sendcount
               k = displs(iam+1) + id
               prjs(:,:,k) = local_prjs(:,:,id)
            enddo
            deallocate(local_prjs)
         endif         

c     the following block of code distributes each process's projections
c     to all other processes using a series of nview_mod broadcast
c     operations. this is probably a little slower than the original
c     method of using a single call to mpi_allgatherv, but avoids the
c     problems associated with mpi buffer limits.

#ifdef MPI
         do i = 0, num_nodes-1
            nprjs = sendcount
            call mpi_bcast(nprjs, 1, mpi_integer, i, mcw, mpi_err)
            do prj_cnt = 1, nprjs
               call mpi_bcast(prjs(1,1,displs(i+1)+prj_cnt),
     $              ncolbin*nrowbin, mpi_real, i, mcw, mpi_err)
            enddo
         enddo
#endif

c     original method for distributing projections
ccc   sendcount  = sendcount  * prj_size
ccc   recvcounts = recvcounts * prj_size
ccc   displs     = displs     * prj_size
ccc   call mpi_allgatherv(prjsl, sendcount,
ccc   +        mpi_real, prjs, recvcounts, displs,
ccc   +        mpi_real, mcw, mpi_err)


      endif

      call system_clock(clock3)

      if(iam == 0) then
         write(*,1000)
         do i=1, nview_mod
            write(*,1010) i, mod_ts(i), mod_ps(i)
         enddo
      endif

 1000 format('  view # theta    phi',/,21('-'),/)
 1010 format('+',i5,f8.2,f7.2)


c  ***********************************
c  *  determine theta,phi,omega,x,y  *
c  ***********************************

      call global_cc (mode, bin, nrot, nannuli, nview_mod,
     1     idim1, idim2, res_min, res_max, 
     2     pixres_min, pixres_max, ann_lo, ann_hi, jcut, radius,
     3     sym, verbose, mag_cen, mag_step, mag_num, 
     4     mod_ts, mod_ps,
     5     prj_avg, prjs, prj, varopt, ncolbin, nrowbin,
     6     nptles, nfiles, img_fnames, param_fnames, ptle_count,
     $     indx, theta, phi, omega, x, y, mag, pft_ccs, 
     8     prj_ccs, cmp_ccs, compare_mode, filt_fac, sigcut,
     9     xcen_model, ycen_model, pftrad_lo, pftrad_hi, quick_omega)

      deallocate(prj_avg, prjs, mod_ts, mod_ps, prj)
C     ----------------------------------------------------

c      call global(ncolbin, nrowbin, nsecbin, nrot, nannuli, nptles,
c     $     nfiles, ptle_count, indx, theta, phi, omega, x, y, mag,
c     $     pft_ccs, prj_ccs, cmp_ccs, compare_mode, filt_fac,
c     $     map_filename)


      call system_clock(clock4) 

      if(iam == 0) then
         if(mode == -1 .or. mode == -2) THEN
            close(3)
            close(4)
         endif

C     Calculate average CCs and MAG factor
         call calc_avgs(nptles, mag, pft_ccs, prj_ccs, cmp_ccs)
         if(verbose >= 1) then
            write(*,*) '=========================================='
            write(*,*) ' Files EMPFT.RADS, EMPFT.RES1, EMPFT.RES2 '
            write(*,*) '  written to default directory.'
            write(*,*) '=========================================='
            close(8)
            close(9)
            close(10)
         endif

         t_proj = elapsed_time(clock2, clock3)
         t_total = elapsed_time(clock1, clock4)
         t_oo = t_total - t_proj

         write(*,*)
         write(*,*) 'Time calculating projections         = ', t_proj
         write(*,*) 'Time determining origins and orients = ', t_oo
         write(*,*) 'Total PPFT run time                  = ', t_total
         write(*,*)

         write(*,*) 'PFT finished.'
      endif

      deallocate(indx, theta, phi, omega, x, y, mag, pft_ccs,
     $     prj_ccs, cmp_ccs)

#ifdef MPI
      call mpi_finalize(mpi_err)
#endif

      end program pftsearch
